@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2000  The PARI group.
+Copyright (C) 2000-2010  The PARI group.
 
 This file is part of the GP2C package.
 
@@ -12,6 +12,7 @@ Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 
+#define TYPE
 #include "config.h"
 #include <stdlib.h>
 #include <stdio.h>
@@ -19,6 +20,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 #include <string.h>
 #include <errno.h>
 #include "header.h"
+
 FILE *yyin;
 int yyparse (void);
 void  yyrestart(FILE *new_file);
@@ -31,6 +33,7 @@ static int optgen=0,opttree=0,opttype=0,optinfo=0;
 int optstrict=0;
 int optcleanvar=0;/*used in genbrace to suppress {} optimization*/
 int indentlevel=2;
+int do_warning=1;
 char *optprefix,*optsuffix;
 void printlistfunc(FILE *fout);
 void outputtype(FILE *fout);
@@ -51,41 +54,101 @@ void dump(FILE *fout)
       {
         if (fc->c[j].flag&(1<<Carg))
           fprintf(fout,"%s%s %s",j?", ":"",
-              Gname[fc->c[j].t],fc->c[j].var);
+              GPname(vartype(fc->c[j])),varstr(fc->c[j]));
       }
       fprintf(fout,")\n");
       if (strcmp(gp->gpname,gp->proto.cname))
         fprintf(fout,"GP name: %s\n",gp->gpname);
       fprintf(fout,"code: %s\n",gp->proto.code);
-      fprintf(fout,"return type: %s\n",Gname[gp->type]);
-      fprintf(fout,"mode=%d \t  spec=%d\n",gp->mode,gp->spec);
+      fprintf(fout,"return type: %s\n",GPname(functype(*gp)));
+      fprintf(fout,"mode=%d \t  spec=%d\n",funcmode(*gp),gp->spec);
       for(   ;j<fc->s.n;j++)
       {
         ctxvar *c=fc->c+j;
         if (!((c->flag&(1<<Carg)) || ((c->flag&(1<<Cconst)) && c->val==-1)))
-          fprintf(fout,"%s %s\n",Gname[c->t],c->var);
+          fprintf(fout,"%s %s\n",GPname(vartype(*c)),varstr(*c));
       }
       fprintf(fout,"\n");
     }
 }
 extern int indent;
-void 
+void
 init_compiler(void)
 {
-  stack_init(&s_node,sizeof(*tree),(void **)&tree);
-  stack_init(&s_value,sizeof(*value),(void **)&value);
-  stack_init(&s_func,sizeof(*lfunc),(void **)&lfunc);
-  stack_init(&s_ctx,sizeof(*ctxstack),(void **)&ctxstack);
-  stack_init(&s_aff,sizeof(*affstack),(void **)&affstack);
-  stack_init(&s_bloc,sizeof(*block),(void **)&block);
-  stack_init(&s_comment,sizeof(*com),(void **)&com);
-  stack_init(&s_errors,sizeof(*errors),(void **)&errors);
+  int n;
+  stack_init(&s_node,sizeof(*tree),(void *)&tree);
+  stack_init(&s_value,sizeof(*value),(void *)&value);
+  stack_init(&s_func,sizeof(*lfunc),(void *)&lfunc);
+  stack_init(&s_ctx,sizeof(*ctxstack),(void *)&ctxstack);
+  stack_init(&s_aff,sizeof(*affstack),(void *)&affstack);
+  stack_init(&s_bloc,sizeof(*block),(void *)&block);
+  stack_init(&s_comment,sizeof(*com),(void *)&com);
+  stack_init(&s_errors,sizeof(*errors),(void *)&errors);
+  stack_init(&s_Ctype,sizeof(*Ctype),(void *)&Ctype);
+  stack_init(&s_GPtype,sizeof(*GPtype),(void *)&GPtype);
+  stack_init(&s_Mmode,sizeof(*Mmode),(void *)&Mmode);
+  stack_init(&s_label,sizeof(*label),(void *)&label);
+   /*Node 0 should be a (Gvoid)Fnoarg*/
+  n=newnode(Fnoarg,-1,-1); tree[n].t=Gvoid;
+   /*Node 1 should be a (Gnotype)Fnoarg*/
+  n=newnode(Fnoarg,-1,-1); tree[n].t=Gnotype;
 }
+
 void
-compile(FILE *fin, char *nom)
+init_stdfunc(void)
+{
+  FC_badtype=findfuncdesc("_badtype");
+  FC_formatcode=findfuncdesc("_formatcode");
+  FC_tovec=findfuncdesc("_tovec");
+  FC_tovecprec=findfuncdescopt("_tovecprec");
+  FC_cast=findfuncdesc("_cast");
+  FC_proto_ret=findfuncdesc("_proto_ret");
+  FC_proto_code=findfuncdesc("_proto_code");
+  FC_decl_base=findfuncdesc("_decl_base");
+  FC_decl_ext=findfuncdesc("_decl_ext");
+  FC_default_check=findfuncdesc("_default_check");
+  FC_default_marker=findfuncdesc("_default_marker");
+  FC_gerepileupto=findfuncdesc("_gerepileupto");
+  FC_const_smallreal=findfuncdesc("_const_smallreal");
+  FC_const_expr=findfuncdesc("_const_expr");
+  FC_copy=findfuncdesc("copy");
+  FC_avma=findfuncdesc("_avma");
+  Gpari_sp=findfunctype("_avma");
+  Gbptr=findfunctype("_diffptr");
+  FC_low_stack_lim=findfuncdesc("_low_stack_lim");
+  /* GP 2.1 has no gerepileall but is supported*/
+  FC_gerepileall=findfuncdescopt("_gerepileall");
+  Ggptr = (FC_gerepileall<0)?strtotype("gptr"):Gnotype;
+  FC_forprime_next=findfuncdescopt("_forprime_next");
+  FC_matrixrow=findfuncdescopt("_[_,]");
+  FC_const_real=findfuncdescopt("_const_real");
+  FC_call=findfuncdescopt("_(_)");
+  FC_strtoclosure=findfuncdescopt("_strtoclosure");
+  FC_proto_max_args=findfuncdescopt("_proto_max_args");
+  max_args = FC_proto_max_args<0 ? 8 :
+             lfunc[FC_proto_max_args].dsc->a[0].args[0].misc;
+}
+
+void
+init_typedef(void)
+{
+  gpdesc *def=lfunc[findfuncdesc("_typedef")].dsc;
+  int r;
+  ctype=(int*)calloc(s_GPtype.n,sizeof(*ctype));
+  for (r=0; r<def->nb; r++)
+  {
+    gpdescarg *rule=def->a+r;
+    int t=rule->args[0].type;
+    ctype[t]=strtoctype(rule->cname);
+  }
+  Vgen=strtoctype("GEN");
+}
+
+void
+compile(FILE *fin, FILE *fout, char *nom)
 {
   char *descfile;
-  int n,startnode;
+  int startnode;
   int status;
   int i;
   init_compiler();
@@ -96,17 +159,17 @@ compile(FILE *fin, char *nom)
   yyin=fin;
   currfunc=-1;
   initoperators();
-  inittype();
   if (!(descfile=getenv("GP2C_FUNC_DSC")))
     descfile=FUNCDSC_PATH;
+  for(i=0;GPneeded[i];i++) newtype(GPneeded[i]);
+  for(i=0; Mneeded[i];i++) newmode( Mneeded[i]);
   initdesc(descfile);
+  inittype();
   patchfunclist();
-  if(opttype) {outputtype(stdout);exit(0);}
-
-  /*Node 0 should be a (Gvoid)Fgnil*/
-  n=newnode(Fgnil,-1,-1); tree[n].t=Gvoid;
-   /*Node 1 should be a (Gnotype)Fgnil*/
-  n=newnode(Fgnil,-1,-1); tree[n].t=Gnotype;
+  if(opttype) {outputtype(fout);exit(0);}
+  init_stdfunc();
+  init_typedef();
+  startnode=s_node.n-1;
   status=yyparse();
   for(i=0;i<s_errors.n;i++)
   {
@@ -123,26 +186,29 @@ compile(FILE *fin, char *nom)
     fprintf(stderr,"%d error%s found: aborting...\n",s_errors.n,s_errors.n==1?"":"s");
     exit(1);
   }
-  if (!s_node.n)
-  {
-    fprintf(stdout,"/*No code was generated*/\n");
-    exit(0);
-  }
-  startnode=s_node.n-1;
+  if (startnode==s_node.n-1)
+    startnode=GNIL;
+  else
+    startnode=s_node.n-1;
   if (optgen)
   {
-    printnode(stdout,startnode);
+    printnode(fout,startnode);
     printf("\n");
     return;
   }
   if (yydebug) fprintf(stderr,"End of parsing\n");
 
-  if (tree[startnode].f!=Fseq && tree[startnode].f!=Fdeffunc)
-    startnode=addseqright(startnode,GNIL);
-  gentoplevel(startnode);
+  startnode=addseqleft(newnode(Ffunction,newentry("_initfunc"),-1),startnode);
+  gentoplevel(startnode,-1);
   startnode=addinitfunc(startnode);
-  if (debug)
+  if (debug>1)
     printnode(stderr,startnode);
+  gentopfunc(startnode,-1,-1,-1);
+  if (debug)
+  {
+    fprintf(stderr,"\n--------------END TOPFUNC------------\n");
+    printnode(stderr,startnode);
+  }
   genblock(startnode,-1);
   if (opttree==1)
   {
@@ -151,19 +217,15 @@ compile(FILE *fin, char *nom)
   }
   if (debug)
     printnode(stderr,startnode);
-  default_type=Gnotype;
-  do
+  lastpass=0;
+  gentype(startnode);
+  do_warning=0;
+  while (lastpass)
   {
     lastpass=0;
     gentype(startnode);
-  }while(lastpass);
-  default_type=Ggen;
-  do
-  {
-    lastpass=0;
-    gentype(startnode);
-  }while(lastpass);
- 
+  }
+  do_warning=1;
   if (debug>=2)
   {
     fprintf(stderr,"\n--------------END GENTYPE------------\n");
@@ -180,17 +242,15 @@ compile(FILE *fin, char *nom)
   {
     lastpass=0;
     varlist(startnode);
-  }while(lastpass);
+  } while(lastpass);
   if (!optcleanvar)
     cleanvar(startnode);
+  pilelist(startnode);
   if (autogc)
-  {
-    pilelist(startnode);
     pileclean(startnode);
-  }
   if (optinfo) dump(stderr);
   if (!optcleanvar)
-      cleanvar(startnode);
+    cleanvar(startnode);
   cleancode(startnode,-1);
   if (tree[startnode].f==Fseq && tree[startnode].y==GNIL)
     startnode=tree[startnode].x;
@@ -209,23 +269,24 @@ compile(FILE *fin, char *nom)
     printnode(stderr,startnode);
   }
   indent=0;
-  genheader(stdout);
-  gencode(stdout,startnode);
+  genheader(fout);
+  gencode(fout,startnode);
 }
 void version(void)
 {
   printf("GP to C compiler version %s \n   targeted at PARI/GP %s\n",VERSION,PARI_VERSION);
-  printf("Copyright 2000-2002 The PARI Group\n");
+  printf("Copyright 2000-2010 The PARI Group\n");
   printf("GP2C is free software, covered by the GNU General Public License, and \n\
 you are welcome to change it and/or distribute copies of it under \n\
 certain conditions.  There is absolutely no warranty for GP2C.\n");
 }
 void usage(FILE *fout, char *s)
 {
-  fprintf(fout,"%s [-ghfltvydSWTGV] [-i N] [-p <prefix>] [file.gp] \n\
+  fprintf(fout,"%s [-ghfltvydSWTGV] [-o <file>] [-i N] [-p <prefix>] [file.gp] \n\
 GP to C translator. \n\
 \n\
 user option: \n\
+ -o <file> :  Place output in file <file>. \n\
  -g : Generate automatic garbage collection code. \n\
  -iN: Set indentation level to N spaces (default 2). \n\
  -W : Output information about global variables. \n\
@@ -247,21 +308,22 @@ debugging options: \n\
  -V : Do not clean up variables. \n\
  \n\
 file.gp: file to be processed, default to stdin. \n\
-The generated C code is output to stdout. \n\
+The generated C code is output to stdout unless the -o option is used \n\
  \n\
 You can automated the process of compiling with \n\
 the script gp2c-run. \n\
-",s); 
+",s);
 }
 int main(int argc, char **argv)
 {
-  char c;
+  int c;
   FILE *fin;
+  FILE *fout=stdout;
   autogc=0;
   warn=0;
   optprefix=NULL;
   optsuffix=NULL;
-  while((c=getopt(argc,argv,"gi:hflp:s:StvydTGVW"))!=-1)
+  while((c=getopt(argc,argv,"gi:hflo:p:s:StvydTGVW"))!=-1)
   {
     switch(c)
     {
@@ -281,14 +343,17 @@ int main(int argc, char **argv)
     case 'l':
       {
         char *descfile;
-        stack_init(&s_func,sizeof(*lfunc),(void **)&lfunc);
+        init_compiler();
         if (!(descfile=getenv("GP2C_FUNC_DSC")))
           descfile=FUNCDSC_PATH;
         initdesc(descfile);
         patchfunclist();
-        printlistfunc(stdout);
+        printlistfunc(fout);
         exit(0);
       }
+    case 'o':
+      fout=fopen(optarg,"w");
+      break;
     case 'p':
       optprefix=strdup(optarg);
       break;
@@ -306,7 +371,7 @@ int main(int argc, char **argv)
       exit(0);
     case 'd':
       debug++;
-      debug_stack=1;
+      debug_stack=debug>1;
       break;
     case 'y':
       yydebug++;
@@ -335,7 +400,7 @@ int main(int argc, char **argv)
     exit(1);
   }
   if (argc==optind)
-    compile(stdin,"stdin");
+    compile(stdin,fout,"stdin");
   else
   {
     char *nom=argv[optind];
@@ -345,8 +410,9 @@ int main(int argc, char **argv)
       perror(argv[0]);
       exit(errno);
     }
-    compile(fin,nom);
+    compile(fin,fout,nom);
     fclose(fin);
   }
+  if (fout!=stdout) fclose(fout);
   return 0;
 }
